import pyfcg as fcg
import json

DEFAULT_GRAMMAR_CONFIGURATION = {"heuristics": [":cxn-score"],
                                 "parse-goal-tests": [":no-applicable-cxns", ":no-sequence-in-root"],
                                 "production-goal-tests": [":no-applicable-cxns", ":no-meaning-in-root"],
                                 "render-mode": ":render-sequences",
                                 "de-render-mode": ":de-render-sequence-predicates",
                                 "draw-meaning-as-network": True,
                                 "cxn-supplier-mode": ":all-cxns",
                                 "construction-inventory-processor-mode": ":heuristic-search",
                                 "node-expansion-mode": ":full-expansion",
                                 "search-algorithm": ":best-first",
                                 "heuristic-value-mode": ":sum-heuristics-and-parent",
                                 "node-tests": [":check-duplicate", ":restrict-nr-of-nodes", ":restrict-search-depth"],
                                 "max-search-depth": 25,
                                 "max-nr-of-nodes": 250,
                                 "category-linking-mode": ":neighbours",
                                 "initial-categorial-link-weight": 0.0}


DEFAULT_GRAMMAR_VISUALIZATION_CONFIGURATION = {"hide-features": ["footprints"],
                                               "show-constructional-dependencies": True}

DEFAULT_FEATURE_TYPES = [['form', 'set-of-predicates', ':handle-regex-sequences'],
                         ['meaning', 'set-of-predicates'],
                         ['form-args', 'sequence'],
                         ['meaning-args', 'sequence'],
                         ['subunits', 'set'],
                         ['args', 'sequence'],
                         ['footprints', 'set']]

DEFAULT_CXN_ATTRIBUTES = {'score': 0.5,
                          'label': 'cxn'}


class Grammar():
    """
    The Grammar class implements all functionality
    required to support FCG grammars in PyFCG.
    A grammar has an id, cxns, a configuration,
    a visualization_configuration and a categorial_network.
    """

    def __init__(self, grammar_id):
        self.id = grammar_id
        self.configuration = DEFAULT_GRAMMAR_CONFIGURATION
        self.visualization_configuration = DEFAULT_GRAMMAR_VISUALIZATION_CONFIGURATION
        self.feature_types = DEFAULT_FEATURE_TYPES
        self.cxns = {}
        self.categorial_network = {}
        self.hashed = False

        self.register_grammar()

    def __repr__(self):
        """String representation for printing construction objects."""
        return "<Grammar: " + str(self.size()) + " constructions>"

    def register_grammar(self):
        """
        Register the grammar in FCG, with all configurations etc. but without constructions
        """
        fcg.routes.register_grammar(self)

    def load_grammar_from_file(self, file_name):
        """Read an FCG grammar (in JSON format) from a file."""
        grammar_spec = {}
        with open(file_name, 'r') as f:
            grammar_spec = json.load(f)
        # Make sure grammar_spec id is equal to Grammar id
        grammar_spec["id"] = self.id
        # Overwrite default settings if user specified configuration,
        # visualization-configuration, categorial-network or feature-types in grammar_spec
        if 'configuration' in grammar_spec:
            self.configuration = grammar_spec['configuration'] #to do: merge configurations!
        if 'visualization-configuration' in grammar_spec:
            self.visualization_configuration = grammar_spec['visualization-configuration'] #to do: merge configurations!
        if 'categorial-network' in grammar_spec:
            self.categorial_network = grammar_spec['categorial-network']
        if 'feature-types' in grammar_spec:
            self.feature_types = grammar_spec['feature-types']

        self.register_grammar()

        cxn_specs = grammar_spec.pop("cxns")
        for cxn_name, cxn_spec in cxn_specs.items():

            if 'feature-types' in cxn_spec:
                feature_types = cxn_spec['feature-types']
            else:
                feature_types = []

            if 'attributes' in cxn_spec:
                attributes = cxn_spec['attributes']
            else:
                attributes = DEFAULT_CXN_ATTRIBUTES

            if 'contributing-pole' in cxn_spec:
                contributing_pole = cxn_spec['contributing-pole']
            else:
                contributing_pole = []


            cxn = fcg.Construction(name=cxn_spec["name"],
                                   conditional_pole=cxn_spec["conditional-pole"],
                                   contributing_pole=contributing_pole,
                                   attributes=attributes,
                                   feature_types=feature_types)
            self.add_cxn(cxn)

    def comprehend(self, utterance):
        return fcg.routes.comprehend(utterance, self.id)

    def comprehend_all(self, utterance):
        comprehension_result = fcg.routes.comprehend_all(utterance, self.id)
        return comprehension_result['meanings'], comprehension_result['applied-cxns']

    def formulate(self, meaning):
        return fcg.routes.formulate(meaning, self.id)

    def formulate_all(self, meaning):
        formulation_result = fcg.routes.formulate_all(meaning, self.id)
        return formulation_result['utterances'], formulation_result['applied-cxns']

    def add_cxn(self, cxn):
        cxn.grammar_id = self.id
        self.cxns[cxn.name] = cxn
        cxn.feature_types = fcg.utils.inherit_feature_types(self.feature_types, cxn.feature_types)
        fcg.routes.add_cxn({"grammar-id" : self.id,
                            "name" : cxn.name,
                            "conditional-pole": cxn.conditional_pole,
                            "contributing-pole": cxn.contributing_pole,
                            "attributes": cxn.attributes,
                            "feature-types": cxn.feature_types})

    def delete_cxn(self, cxn):
        if type(cxn) is str:
            self.cxns.pop(cxn)
            fcg.routes.delete_cxn(cxn, self.id)
        else:
            self.cxns.pop(cxn.name)
            fcg.routes.delete_cxn(cxn.name, self.id)

    def size(self):
        return len(self.cxns)

    def clear_cxns(self):
        self.cxns={}
        fcg.routes.clear_grammar(self.id)

    def find_cxn_by_name (self, cxn_name):
        return self.cxns[cxn_name]

    def show_in_web_interface (self):
        """Show grammar in FCG's web interface."""
        lisp_expression = '(add-element (make-html (eval ' + self.id  + ') :expand-initially t))'
        fcg.routes.lisp_eval(lisp_expression, package='fcg-go')

    def add_category(self, category):
        lisp_expression = "(add-category '" + category + " (eval " + self.id  + "))"
        fcg.routes.lisp_eval(lisp_expression, package='fcg-go')
        self.categorial_network['nodes'].append(category)

    def add_link(self, category_1, category_2):
        lisp_expression = "(add-link '" + category_1 + " '" + category_2 + " (eval " + self.id + "))"
        fcg.routes.lisp_eval(lisp_expression, package='fcg-go')
        self.categorial_network['edges'].append([category_1, category_2])