import requests
from requests import JSONDecodeError

import pyfcg.fcg_go_bridge as fcg_go

endpoints = {'test_connection': '/fcg-test-connection',
             'lisp_eval': '/fcg-read-eval-print',
             'start_web_interface': "/fcg-start-web-interface",
             'activate_monitors': "/fcg-activate-monitors",
             'load_demo_grammar': "/fcg-load-demo-grammar",
             'comprehend': "/fcg-comprehend",
             'comprehend_all': "/fcg-comprehend-all",
             'formulate': "/fcg-formulate",
             'formulate_all': "/fcg-formulate-all",
             'add_cxn': "/fcg-add-cxn",
             'delete_cxn': "/fcg-delete-cxn",
             'register_grammar': "/fcg-initialise-grammar",
             'clear_grammar': "/fcg-clear-grammar",
             'deactivate_all_monitors': "/fcg-deactivate-all-monitors",
             'fcg_eval': "/fcg-apply-fn",
             'set_cxn_score': "/fcg-set-cxn-score",
             'gensym': "/fcg-gensym"}

# Get and post requests #
#########################

class FcgError(Exception):
    pass


def get (endpoint, params=None):
    response = requests.get('http://' + fcg_go.server_address + ':' + str(fcg_go.server_port) + endpoint,params=params)
    if response.status_code == 200:
        return response.json()
    else:
        try:
            error_message = response.json()['error-message']
        except JSONDecodeError:
            error_message = response.text
        raise FcgError(error_message)


def post (endpoint, data, params=None):
    response = requests.post('http://' + fcg_go.server_address + ':' + str(fcg_go.server_port) + endpoint, json=data, params=params)
    if response.status_code == 200:
        return response.json()
    else:
        try:
            error_message = response.json()['error-message']
        except JSONDecodeError:
            error_message = response.text
        raise FcgError(error_message)

# System and testing #
#########################

def test_connection ():
    endpoint = endpoints['test_connection']
    return get(endpoint)


def lisp_eval (lisp_expression, package=None):
    """Execute arbitrary code in package."""
    endpoint = endpoints['lisp_eval']
    data = {'lisp-code' : lisp_expression, 'package': package}
    return post(endpoint, data)

def fcg_eval (*args,package=None):
    """Execute arbitrary code in package."""
    endpoint = endpoints['fcg_eval']
    data = {'function-name' : args[0],
            'args': args[1:],
            'package': package}
    return post(endpoint, data)

def gensym(string,variable=False,package=None):
    endpoint = endpoints['gensym']
    data = {'string' : string,
            'variable' : variable,
            'package': package}
    return post(endpoint, data)



# Web interface and monitors #
##############################

def start_web_interface (address='localhost', port=8010, open=True):
    """Start up the babel web interface."""
    endpoint = endpoints['start_web_interface']
    data = {'address': address, 'port': port, 'open': open}
    return post(endpoint, data)

def activate_monitors (list_of_monitors):
    """Start up the babel web interface."""
    endpoint = endpoints['activate_monitors']
    data = {'monitors': list_of_monitors}
    return post(endpoint, data)

def deactivate_all_monitors ():
    """Start up the babel web interface."""
    endpoint = endpoints['deactivate_all_monitors']
    return get(endpoint)


# Demo grammar #
################

def load_demo_grammar ():
    """Load an FCG demo grammar."""
    endpoint = endpoints['load_demo_grammar']
    return get(endpoint)


# FCG processing #
##################

def comprehend(utterance, grammar_id='*fcg-constructions*', package =None):
    endpoint = endpoints['comprehend']
    data = {'utterance': utterance,
            'grammar': grammar_id,
            'package': package}
    return post(endpoint, data)

def comprehend_all(utterance, grammar_id='*fcg-constructions*', package=None):
    endpoint = endpoints['comprehend_all']
    data = {'utterance': utterance,
            'grammar': grammar_id,
            'package':package}
    return post(endpoint, data)

def formulate(meaning, grammar_id='*fcg-constructions*', package=None):
    endpoint = endpoints['formulate']
    data = {'meaning': meaning,
            'grammar': grammar_id,
            'package':package}
    return post(endpoint, data)


def formulate_all(meaning, grammar_id='*fcg-constructions*', package=None):
    endpoint = endpoints['formulate_all']
    data = {'meaning': meaning,
            'grammar': grammar_id,
            'package':package}
    return post(endpoint, data)

# Grammars and constructions #
##############################

def add_cxn(cxn_spec):
    endpoint = endpoints['add_cxn']
    data = {'cxn': cxn_spec}
    return post(endpoint, data)

def delete_cxn(cxn_name, grammar_id):
    endpoint = endpoints['delete_cxn']
    data = {'cxn-name': cxn_name,
            'grammar': grammar_id}
    return post(endpoint, data)

def register_grammar(grammar):
    """Initialises an FCG grammar (without cxns) based on a grammar specification."""
    endpoint = endpoints['register_grammar']
    data = {'hashed' : grammar.hashed,
            'id': grammar.id,
            'feature-types': grammar.feature_types,
            'configuration': grammar.configuration,
            'visualization-configuration': grammar.visualization_configuration,
            'categorial-network': grammar.categorial_network}
    return post(endpoint, data)

def clear_grammar(grammar_id):
    endpoint = endpoints['clear_grammar']
    data = {'grammar': grammar_id}
    return post(endpoint, data)

def set_cxn_score(cxn_name, new_score, grammar_id='*fcg-constructions*'):
    endpoint = endpoints['set_cxn_score']
    data = {'cxn-name': cxn_name,
            'new-score': new_score,
            'grammar': grammar_id}
    return post(endpoint, data)

